import itertools
from collections import defaultdict
from typing import List, Dict, Optional

from server.model.card import Card
from server.model.deck import Deck
from server.model.hand import Hand
from server.model.players import Player


class Game:
    def __init__(self, players=None, deck: Deck = Deck()):
        if players is None:
            players = []

        self.deck: Deck = deck
        self.players: List[Player] = players
        self.defeats: Dict[Player, int] = defaultdict(lambda: 0)
        self.current_bet: Optional[Hand] = None

    @property
    def global_hand(self) -> Hand:
        return Hand([c for p in self.players for c in p.hand])

    def new_game(self):
        self.deck.reset()

        while len(self.players) > 1:
            loser = self.new_turn()

            if self.defeats[loser] == 5:
                print(f"{loser} is eliminated!")
                self.players.remove(loser)
            else:
                print(f"{loser} lost the round, now playing with {self.defeats[loser] + 1} cards!")
                self.players.remove(loser)
                self.players.insert(0, loser)

        winner = self.players[0]
        print(f"Game over - {winner.name} wins with {len(winner.hand)} cards!")

    def new_turn(self) -> Player:
        """
        Runs a turn of the game.

        :return: the player that lost this turn.
        """
        # Distribution
        self.deck.reset()
        for current_player in self.players:
            current_player.clear()
            draw = []
            for i in range(self.defeats[current_player] + 1):
                draw.append(self.deck.random_card())
            print(f"Drew {draw} for {current_player}.")
            for d in draw:
                current_player.give(d)

        # Tour
        self.current_bet = None
        last_player = None
        for current_player in itertools.cycle(self.players):
            loser = self.play_turn(current_player, last_player)
            if loser is not None:
                return loser
            last_player = current_player

        # TODO: Put next first player first of list

    def is_menteur(self, bet: Hand):
        """
        Is this bet a menteur?
        :param bet:
        :return:
        """

        to_scan = [Card(c.value) for c in self.global_hand.cards.copy()]  # Keep only values
        to_find = [Card(c.value) for c in bet.cards]
        menteur = False

        for card in to_find:
            if card in to_scan:
                to_scan.remove(card)
                continue
            else:
                print(f"Missing {card}!")
                menteur = True
        if menteur:
            print(f"Didn't find {bet} in {self.global_hand}: MENTEUR!")
        return menteur

    def count_defeat(self, loser):
        self.defeats[loser] += 1

    def add_player(self, player: Player):
        """
        Adds a player to the game.
        :return: True if the player was accepted (unique names).
        """
        if any([s.name == player.name for s in self.players]):
            return False
        else:
            self.players.append(player)
            return True

    def play_turn(self, current_player: Player, last_player: Player) -> Optional[Player]:
        """
        Runs a turn, eventually returning a loser.

        :param current_player:
        :param last_player:
        :return:
        """
        print(f"| {current_player}'s turn >")

        if not self.current_bet:  # First player, has to bet something
            while not self.current_bet:  # Ask a valid bet
                announce = current_player.announce(self.current_bet)
                if announce.bet:
                    self.current_bet = announce.bet
                    print(f"{current_player} starts the round: {self.current_bet}")

        else:  # Next player, announce or menteur
            announce = current_player.announce(self.current_bet)

            if announce.bet:
                while announce.bet.value() < self.current_bet.value():  # Bad announce!
                    print(f"Invalid bet by {current_player}, {announce.bet} < {self.current_bet}!")
                    announce = current_player.announce(self.current_bet)
                self.current_bet = announce.bet

                # Valid bet:
                print(f" {current_player} bets {self.current_bet}.")

            else:  # Menteur! Who lost the round?
                print(f" {current_player} says Menteur!")
                menteur = self.is_menteur(self.current_bet)
                loser = last_player if menteur else current_player
                self.count_defeat(loser)

                print(f"{current_player} says Menteur... {'Bravo!' if menteur else 'FAIL!'}")
                return loser
